# Copyright 2023 Euratom
# Copyright 2023 United Kingdom Atomic Energy Authority
# Copyright 2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.
"""Module for loading edge plasma profiles from the edge_profiles IDS."""

from dataclasses import fields

import numpy as np
from numpy.typing import NDArray
from raysect.core.math import Vector3D, translate
from raysect.core.math.function.float import Constant2D, Constant3D, Function2D
from raysect.core.math.function.vector3d import Constant2D as ConstantVector2D
from raysect.core.math.function.vector3d import Constant3D as ConstantVector3D
from raysect.core.math.function.vector3d import Function2D as VectorFunction2D
from raysect.core.scenegraph._nodebase import _NodeBase
from raysect.primitive import Cylinder, Subtract
from scipy.constants import atomic_mass, electron_mass

from cherab.core import AtomicData, Maxwellian, Plasma, Species
from cherab.core.math import AxisymmetricMapper, VectorAxisymmetricMapper
from cherab.tools.equilibrium.efit import FluxSurfaceNormal, PoloidalFieldVector
from imas import DBEntry
from imas.ids_structure import IDSStructure

from ..ggd.base_mesh import GGDGrid
from ..ids.common import get_ids_time_slice
from ..ids.common.ggd import load_grid
from ..ids.common.species import ProfileData, VelocityData
from ..ids.edge_profiles import load_edge_species
from ..math import UnitVector2D
from .equilibrium import load_equilibrium, load_magnetic_field
from .utility import (
    ZERO_VELOCITY,
    ProfileInterpolator,
    get_subset_name_index,
    warn_unsupported_species,
)

__all__ = ["load_edge_plasma"]


def load_edge_plasma(
    *args,
    time: float = 0,
    occurrence: int = 0,
    grid_ggd: IDSStructure | None = None,
    grid_subset_id: int | str = 5,
    b_field: VectorFunction2D | None = None,
    time_threshold: float = np.inf,
    split_ion_bundles: bool = True,
    atomic_data: AtomicData | None = None,
    parent: _NodeBase | None = None,
    **kwargs,
) -> Plasma:
    """Load edge profiles and create a `~cherab.core.plasma.node.Plasma` object.

    Prefer ``density_thermal`` over ``density`` profile.

    The distribution of each species is defined with `~cherab.core.distribution.Maxwellian` using
    its density, temperature, and bulk velocity profiles, which are mapped to GGD grid coordinates.

    The plasma geometry is defined as a cylindrical annulus between the inner and outer limits of
    the grid, and its height is defined as the difference between the maximum and minimum
    z-coordinates of the grid.

    The ion bundle species are split into their constituent charge states
    using `.solve_coronal_equilibrium` when `.split_ion_bundles` is True and the necessary atomic
    data is available. Otherwise, ion bundles are ignored with a warning.

    Parameters
    ----------
    *args
        Arguments passed to the `~imas.db_entry.DBEntry` constructor.
    time
        Time for the edge plasma, by default 0.
    occurrence
        Occurrence index of the ``edge_profiles`` IDS, by default 0.
    grid_ggd
        Alternative ``grid_ggd`` structure with the grid description, by default None.
    grid_subset_id
        Identifier of the grid subset. Either index or name, by default 5 (``"Cells"``).
    b_field
        Alternative 2D interpolator of the magnetic field vector (Br, Bphi, Bz).
        Default is None. The magnetic field will be loaded from the ``equilibrium`` IDS.
    time_threshold
        Maximum allowed difference between the requested time and the nearest
        available time, by default `numpy.inf`.
    split_ion_bundles
        Whether to split ion bundles into their constituent charge states using
        `.solve_coronal_equilibrium`, by default True.
    atomic_data
        Atomic data provider class for this plasma, by default None.
        If None, some species (e.g. ion_bundle) may not be properly loaded.
    parent
        Parent node in the Raysect scene graph, by default None.
        Typically a `~raysect.optical.scenegraph.world.World` instance.
    **kwargs
        Keyword arguments passed to the `~imas.db_entry.DBEntry` constructor.

    Returns
    -------
    `~cherab.core.plasma.node.Plasma`
        Plasma object with edge profiles.

    Raises
    ------
    RuntimeError
        If the ``grid_ggd`` or ``ggd`` AOS of the edge_profiles IDS is empty.
    """
    # -----------------------------------------
    # === Load edge profiles data from IMAS ===
    # -----------------------------------------
    # Load required data from the edge_profiles IDS and form the edge grid and species composition
    # data structures.
    with DBEntry(*args, **kwargs) as entry:
        edge_profiles_ids = get_ids_time_slice(
            entry, "edge_profiles", time=time, occurrence=occurrence, time_threshold=time_threshold
        )

    if not len(edge_profiles_ids.grid_ggd) and grid_ggd is None:
        raise RuntimeError(
            "The 'grid_ggd' AOS of the edge_profiles IDS is empty "
            + "and an alternative grid_ggd structure is not provided."
        )

    if not len(edge_profiles_ids.ggd):
        raise RuntimeError("The 'ggd' AOS of the edge_profiles IDS is empty.")

    # Load magnetic field data. If not provided, try to load from the equilibrium IDS.
    if b_field is None:
        try:
            b_field = load_magnetic_field(*args, time=time, occurrence=occurrence, **kwargs)
        except RuntimeError:
            try:
                b_field = load_equilibrium(
                    *args, time=time, occurrence=occurrence, **kwargs
                ).b_field
            except RuntimeError:
                print("Warning! No magnetic field data available in the equilibrium IDS.")

    # Create edge grid
    grid_ggd = grid_ggd or edge_profiles_ids.grid_ggd[0]
    grid, subsets, subset_id = load_grid(grid_ggd, with_subsets=True)

    grid_subset_name, grid_subset_index = get_subset_name_index(subset_id, grid_subset_id)

    if np.all(subsets[grid_subset_name] != np.arange(grid.num_cell, dtype=int)):
        # To reduce memory usage, create the sub-grid only if needed.
        grid = grid.subset(subsets[grid_subset_name], name=grid_subset_name)

    # Load species composition
    composition = load_edge_species(
        edge_profiles_ids.ggd[0],
        grid_subset_index=grid_subset_index,
        split_ion_bundles=split_ion_bundles,
        atomic_data=atomic_data,
    )

    # ----------------------------
    # === Create Plasma object ===
    # ----------------------------
    name = f"IMAS edge plasma: time {edge_profiles_ids.time[0]}, uri {entry.uri}."
    plasma = Plasma(parent=parent, name=name)

    # Create plasma geometry
    radius_outer = grid.mesh_extent["rmax"]
    radius_inner = grid.mesh_extent["rmin"]
    height = grid.mesh_extent["zmax"] - grid.mesh_extent["zmin"]
    zmin = grid.mesh_extent["zmin"]
    plasma.geometry = Subtract(Cylinder(radius_outer, height), Cylinder(radius_inner, height))
    plasma.geometry_transform = translate(0, 0, zmin)

    # Add magnetic field
    if b_field is not None:
        plasma.b_field = VectorAxisymmetricMapper(b_field)

    # Add atomic data
    if atomic_data is not None:
        plasma.atomic_data = atomic_data

    # ------------------------------------
    # === Define Electron Distribution ===
    # ------------------------------------
    if composition.electron.density_thermal is not None:
        composition.electron.density = composition.electron.density_thermal

    if composition.electron is None:
        raise RuntimeError("Electron profile is missing in the edge_profiles IDS.")
    if composition.electron.density is None:
        raise RuntimeError("Electron density profile is missing in the edge_profiles IDS.")
    if composition.electron.temperature is None:
        raise RuntimeError("Electron temperature profile is missing in the edge_profiles IDS.")

    interp = get_edge_interpolators(grid, composition.electron, b_field, return3d=True)

    plasma.electron_distribution = Maxwellian(
        interp.density,
        interp.temperature,
        interp.velocity or ZERO_VELOCITY,
        electron_mass,
    )

    # -----------------------------------------------
    # === Define Species Distribution/Composition ===
    # -----------------------------------------------

    # === Ion/Neutral Species ===
    for profile in composition.ion + composition.neutral:
        if profile.species is None:
            print(f"Warning! Skipping species with missing element or charge: {profile}")
            continue
        if profile.species.element is None:
            print(f"Warning! Skipping species with missing element: {profile}")
            continue
        if profile.density_thermal is not None:
            profile.density = profile.density_thermal
        if profile.density is None:
            print(f"Warning! Skipping {profile.species}: density profile is missing.")
            continue
        if profile.temperature is None:
            print(f"Warning! Skipping {profile.species}: temperature profile is missing.")
            continue

        element = profile.species.element
        charge = profile.species.z_min

        try:
            species = plasma.composition.get(element, int(charge))
            print(f"Warning! Skipping {species}: already defined")
            continue
        except ValueError:
            pass

        interp = get_edge_interpolators(grid, profile, b_field, return3d=True)

        distribution = Maxwellian(
            interp.density,
            interp.temperature,
            interp.velocity or ZERO_VELOCITY,
            element.atomic_weight * atomic_mass,
        )

        plasma.composition.add(Species(element, int(charge), distribution))

    # === Ion Bundles ===
    # Ion bundles are split into their constituent charge states at the composition level.
    warn_unsupported_species(composition, "ion_bundle")

    # === Molecular Species ===
    # TODO: properly support molecular species.
    # For now, just issue a warning if any molecule or molecular_bundle species are present in the composition.
    warn_unsupported_species(composition, "molecule")
    warn_unsupported_species(composition, "molecular_bundle")

    return plasma


def get_edge_interpolators(
    grid: GGDGrid,
    profile: ProfileData,
    b_field: VectorFunction2D | None = None,
    return3d: bool = False,
) -> ProfileInterpolator:
    """Create interpolators for the profiles defined on a grid.

    Parameters
    ----------
    grid
        GGD-compatible grid object.
    profile
        Instance of the `ProfileData` dataclass containing the profiles to be interpolated.
    b_field
        2D interpolator of the magnetic field vector (Br, Btor, Bz), by default None.
    return3d
        If True, return the 3D functions for 2D interpolators assuming
        rotational symmetry, by default False.

    Returns
    -------
    ProfileInterpolator
        Instance of the `ProfileInterpolator` dataclass containing the interpolators for density and temperature.
    """
    interpolators = ProfileInterpolator()

    for field in fields(profile):
        if field.name in {"species", "velocity"}:
            continue
        data = getattr(profile, field.name, None)
        if data is not None:
            func = grid.interpolator(data)
            if isinstance(func, Function2D) and return3d:
                func = AxisymmetricMapper(func)
            setattr(interpolators, field.name, func)

    # Create velocity interpolator
    if profile.velocity is not None:
        vector_func = _get_velocity_interpolators(grid, profile.velocity, b_field)
        if isinstance(vector_func, VectorFunction2D) and return3d:
            vector_func = VectorAxisymmetricMapper(vector_func)
        interpolators.velocity = vector_func

    return interpolators


def _get_velocity_interpolators(
    grid: GGDGrid, v: VelocityData, b_field: VectorFunction2D | None = None
):
    if b_field is None:
        return _get_cylindrical_velocity_interpolators(grid, v.r, v.z, v.phi)

    if v.radial is None and v.r is not None and v.z is not None:
        v_phi = v.phi
        if v_phi is None and v.parallel is not None:
            _, v_phi = _get_components_from_vpar(grid, v.parallel, b_field)
        return _get_cylindrical_velocity_interpolators(grid, v.r, v.z, v_phi)

    if v.parallel is None:
        return _get_poloidal_velocity_interpolators(grid, v.poloidal, v.radial, v.phi, b_field)

    return _get_parallel_velocity_interpolators(grid, v.parallel, v.radial, b_field)


def _get_cylindrical_velocity_interpolators(
    grid: GGDGrid,
    vr: NDArray[np.float64] | None,
    vz: NDArray[np.float64] | None,
    vtor: NDArray[np.float64] | None,
):
    if vr is None and vz is None and vtor is None:
        if grid.dimension == 2:
            return ConstantVector2D(
                Vector3D(0, 1.0e-16, 0)
            )  # avoid zero-length vectors for blending

        return ConstantVector3D(Vector3D(0, 1.0e-16, 0))  # avoid zero-length vectors for blending

    if vr is None:
        vr = np.zeros(grid.num_cell, dtype=np.float64)
    if vz is None:
        vz = np.zeros(grid.num_cell, dtype=np.float64)
    if vtor is None:
        vtor = np.zeros(grid.num_cell, dtype=np.float64)

    return grid.vector_interpolator(np.array([vr, vtor, vz]))


def _get_parallel_velocity_interpolators(
    grid: GGDGrid,
    vpar: NDArray[np.float64] | None,
    vrad: NDArray[np.float64] | None,
    b_field: VectorFunction2D,
):
    if vpar is None and vrad is None:
        if grid.dimension == 2:  # 2D case
            return ConstantVector2D(
                Vector3D(0, 1.0e-16, 0)
            )  # avoid zero-length vectors for blending

        return ConstantVector3D(Vector3D(0, 1.0e-16, 0))  # avoid zero-length vectors for blending

    const_func = Constant2D if grid.dimension == 2 else Constant3D

    vpar_i = const_func(0) if vpar is None else grid.interpolator(vpar)
    vrad_i = const_func(0) if vrad is None else grid.interpolator(vrad)

    parallel_vector = UnitVector2D(b_field)
    surface_normal = FluxSurfaceNormal(b_field)

    if grid.dimension == 3:  # 3D case
        parallel_vector = VectorAxisymmetricMapper(parallel_vector)
        surface_normal = VectorAxisymmetricMapper(surface_normal)

    return vpar_i * parallel_vector + vrad_i * surface_normal


def _get_poloidal_velocity_interpolators(
    grid: GGDGrid,
    vpol: NDArray[np.float64] | None,
    vrad: NDArray[np.float64] | None,
    vtor: NDArray[np.float64] | None,
    b_field: VectorFunction2D,
):
    if vpol is None and vrad is None and vtor is None:
        if grid.dimension == 2:  # 2D case
            return ConstantVector2D(
                Vector3D(0, 1.0e-16, 0)
            )  # avoid zero-length vectors for blending

        return ConstantVector3D(Vector3D(0, 1.0e-16, 0))  # avoid zero-length vectors for blending

    const_func = Constant2D if grid.dimension == 2 else Constant3D

    vpol_i = const_func(0) if vpol is None else grid.interpolator(vpol)
    vrad_i = const_func(0) if vrad is None else grid.interpolator(vrad)
    vtor_i = const_func(0) if vtor is None else grid.interpolator(vtor)

    poloidal_vector = PoloidalFieldVector(b_field)
    surface_normal = FluxSurfaceNormal(b_field)
    toroidal_vector = ConstantVector2D(Vector3D(0, 1, 0))

    if grid.dimension == 3:  # 3D case
        poloidal_vector = VectorAxisymmetricMapper(poloidal_vector)
        surface_normal = VectorAxisymmetricMapper(surface_normal)
        toroidal_vector = VectorAxisymmetricMapper(toroidal_vector)

    return vpol_i * poloidal_vector + vrad_i * surface_normal + vtor_i * toroidal_vector


def _get_components_from_vpar(grid: GGDGrid, vpar: NDArray[np.float64], b_field: VectorFunction2D):
    vpol = np.zeros(grid.num_cell, dtype=np.float64)
    vtor = np.zeros(grid.num_cell, dtype=np.float64)

    for i, cell_centre in enumerate(grid.cell_centre):
        if grid.dimension == 2:  # 2D case
            r, z = cell_centre
        else:  # 3D case
            if grid.coordinate_system == "cartesian":
                r = np.sqrt(cell_centre[0] ** 2 + cell_centre[1] ** 2)
                z = cell_centre[2]
            else:
                r, _, z = cell_centre
        try:
            b_field = b_field(r, z)
            vpol[i] = np.sqrt(b_field.x**2 + b_field.z**2) * (vpar[i] / b_field.length)
            vtor[i] = vpar[i] * b_field.y / b_field.length
        except ValueError:  # Outside equilibrium grid
            continue

    return vpol, vtor



# Copyright 2023 Euratom
# Copyright 2023 United Kingdom Atomic Energy Authority
# Copyright 2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.
"""Module for loading core-profile-related data from IMAS IDS structures."""

from dataclasses import astuple, dataclass, fields

import numpy as np
from numpy.typing import NDArray

from cherab.core.atomic import AtomicData
from imas.ids_primitive import IDSNumericArray
from imas.ids_structure import IDSStructure

from ..common import solve_coronal_equilibrium
from ..common.species import (
    ProfileData,
    SpeciesComposition,
    SpeciesData,
    SpeciesType,
    VelocityData,
    get_elements,
    get_ion,
    get_ion_state,
    get_neutral,
    get_neutral_state,
)

__all__ = [
    "GridData",
    "load_core_grid",
    "load_core_profiles",
    "load_core_species",
]


@dataclass
class GridData:
    """Dataclass for storing grid properties of the core profiles."""

    rho_tor_norm: NDArray[np.float64] | None = None
    """Normalized toroidal flux coordinate."""
    psi: NDArray[np.float64] | None = None
    """Toroidal flux [Wb]."""
    volume: NDArray[np.float64] | None = None
    """Volume enclosed by the flux surface [m^3]."""
    area: NDArray[np.float64] | None = None
    """Area of the flux surface [m^2]."""
    surface: NDArray[np.float64] | None = None
    """Surface-averaged value of the profile on the flux surface."""


def _get_profile(ids_struct: IDSStructure, name: str, name2: str | None = None):
    data = getattr(ids_struct, name, None)
    if isinstance(data, IDSNumericArray):
        if len(data):
            return np.asarray(data)
        else:
            return None
    elif isinstance(data, IDSStructure):
        # Try to search lower-level structure
        return _get_profile(data, name2) if name2 is not None else None
    else:
        return None


def load_core_grid(grid_struct: IDSStructure) -> GridData:
    """Load grid properties of the core profiles.

    The returned dictionary values for missing data are None.

    Parameters
    ----------
    grid_struct
        The IDS structure containing the grid data for 1D profiles.

    Returns
    -------
    `.GridData`
        Instance of the `.GridData` dataclass containing the grid properties for the core profiles.
    """
    grid = GridData()
    for field in fields(grid):
        setattr(grid, field.name, _get_profile(grid_struct, field.name))

    return grid


def load_core_profiles(
    species_struct: IDSStructure,
    species: SpeciesData,
    backup_species_struct: IDSStructure | None = None,
) -> ProfileData:
    """Load core profiles from a given species structure.

    Parameters
    ----------
    species_struct
        IDS structure containing the profiles for a single species.
    species
        The species data for the profiles to be loaded.
    backup_species_struct
        The backup ids structure that is used if the profile is missing in species_struct,
        by default None.

    Returns
    -------
    `.ProfileData`
        Instance of the `ProfileData` dataclass containing the core profiles for the species.
    """
    profiles = ProfileData(species)
    velocities = VelocityData()

    for field in fields(profiles):
        match field.name:
            case "velocity":
                for field_vel in fields(velocities):
                    data = _get_profile(species_struct, field.name, name2=field_vel.name)
                    if data is None and backup_species_struct is not None:
                        data = _get_profile(backup_species_struct, field.name, name2=field_vel.name)
                    setattr(velocities, field_vel.name, data)
                profiles.velocity = velocities
            case "species":
                # species data is not stored in the profile structure, skip loading it here
                continue
            case _:
                setattr(profiles, field.name, _get_profile(species_struct, field.name))
                if getattr(profiles, field.name) is None and backup_species_struct is not None:
                    setattr(profiles, field.name, _get_profile(backup_species_struct, field.name))

    return profiles


def load_core_species(
    profile_1d: IDSStructure,
    split_ion_bundles: bool = True,
    atomic_data: AtomicData | None = None,
) -> SpeciesComposition:
    """Load core plasma species and their profiles from a given profiles IDS structure.

    Parameters
    ----------
    profile_1d
        The IDS structure containing the core profiles data.
    split_ion_bundles
        Whether to split ion bundles into individual ion states using `.solve_coronal_equilibrium`,
        by default True.
    atomic_data
        Optional atomic data to pass to `.solve_coronal_equilibrium` when splitting ion bundles.

    Returns
    -------
    `.SpeciesComposition`
        Instance of the `.SpeciesComposition` dataclass

    Raises
    ------
    RuntimeError
        If electron temperature or density profiles are missing, which are required for determining
        the species composition and solving coronal equilibrium when splitting ion bundles.
    RuntimeError
        If unable to determine the species due to missing element information, density profiles,
        or other necessary data.
    """
    composition = SpeciesComposition(
        electron=load_core_profiles(profile_1d.electrons, SpeciesData(z_min=-1, z_max=-1)),
    )

    if composition.electron.temperature is None:
        raise RuntimeError("Electron temperature profiles are required.")

    electron_density = composition.electron.density_thermal
    if electron_density is None:
        electron_density = composition.electron.density
    if electron_density is None:
        raise RuntimeError("Electron density or density_thermal profiles are required.")

    # Temporary sets
    ion_elements = []
    ion_uuids = set()
    ion_bundle_uuids = set()
    neutral_uuids = set()
    neutral_bundle_uuids = set()

    # ------------
    # === Ions ===
    # ------------
    for ion in profile_1d.ion:
        ion: IDSStructure
        elements = get_elements(ion.element)
        if not len(elements):
            raise RuntimeError("Unable to determine the ion species, ion.element AOS is empty.")
        ion_elements.append(elements)

        # ---------------
        # === Bundles ===
        # ---------------
        if len(ion.state):
            shared_temperature = _get_profile(ion, "temperature")

            # Use the ion-level profiles as backup if there is only one state, otherwise no backup
            backup_ids = None if len(ion.state) > 1 else ion

            for i, state in enumerate(ion.state):
                species_data = get_ion_state(state, i, elements)
                species_tuple = astuple(species_data)
                if species_tuple in ion_bundle_uuids or species_tuple in ion_uuids:
                    print(f"Warning! Skipping duplicated ion: {species_data}")
                    continue

                profile_data = load_core_profiles(
                    state,
                    species_data,
                    backup_species_struct=backup_ids,
                )

                if backup_ids is None and profile_data.temperature is None:
                    profile_data.temperature = shared_temperature

                # === Case: ION ===
                if species_data.species_type == SpeciesType.ION:
                    composition.ion.append(profile_data)
                    ion_uuids.add(species_tuple)

                # === Case: ION_BUNDLE ===
                elif species_data.species_type == SpeciesType.ION_BUNDLE:
                    # Split ion bundles into individual ion states
                    if split_ion_bundles:
                        # Check if the necessary data is available before attempting to solve coronal equilibrium
                        elm = species_data.element or species_data.elements[0]
                        if elm is None:
                            raise RuntimeError(
                                f"Unable to determine the element for ion {species_data}, "
                                "cannot solve coronal equilibrium to split ion bundle."
                            )
                        density = profile_data.density_thermal
                        if density is None:
                            density = profile_data.density
                        if density is None:
                            raise RuntimeError(
                                f"Missing density profiles for ion {species_data}, "
                                "cannot solve coronal equilibrium to split ion bundle."
                            )
                        try:
                            densities_per_charge = solve_coronal_equilibrium(
                                elm,
                                density,
                                electron_density,
                                composition.electron.temperature,
                                atomic_data=atomic_data,
                                z_min=species_data.z_min,
                                z_max=species_data.z_max,
                            )
                        except Exception as e:
                            print(
                                f"Skipping ion bundle {species_data} "
                                f"due to error in solving coronal equilibrium: {e}"
                            )
                            composition.ion_bundle.append(profile_data)
                            ion_bundle_uuids.add(astuple(species_data))
                            continue

                        charge_states = np.arange(
                            species_data.z_min, species_data.z_max + 1, dtype=int
                        )

                        for i_charge, charge in enumerate(charge_states):
                            species = SpeciesData(
                                element=species_data.element,
                                z_min=charge,
                                z_max=charge,
                                species_type=SpeciesType.ION,
                            )
                            composition.ion.append(
                                ProfileData(
                                    species=species,
                                    density=densities_per_charge[i_charge],
                                    density_thermal=densities_per_charge[i_charge],
                                    temperature=profile_data.temperature,
                                    velocity=profile_data.velocity,
                                )
                            )
                            ion_uuids.add(astuple(species))

                    # Don't split ion bundles, just add the bundle as is
                    else:
                        composition.ion_bundle.append(profile_data)
                        ion_bundle_uuids.add(astuple(species_data))

                # === Case: Unexpected species type ===
                else:
                    print(f"Warning! Skipping ion with unexpected species {species_data}")

        # -------------------
        # === Non-bundled ===
        # -------------------
        else:
            species_data = get_ion(ion, elements)
            species_tuple = astuple(species_data)
            if species_tuple in ion_uuids:
                print(f"Warning! Skipping duplicated ion: {species_data}")
            else:
                profile_data = load_core_profiles(ion, species_data)
                if species_data.species_type == SpeciesType.ION:
                    composition.ion.append(profile_data)
                    ion_uuids.add(species_tuple)
                else:
                    print(
                        f"Warning! Skipping non-bundled ion with unexpected species {species_data}"
                    )

    # ----------------------------
    # === Neutrals (molecules) ===
    # ----------------------------
    for neutral in profile_1d.neutral:
        elements = get_elements(neutral.element)
        if not len(elements):
            elements = ion_elements[neutral.ion_index - 1]

        # ---------------
        # === Bundles ===
        # ---------------
        if len(neutral.state):
            shared_temperature = _get_profile(neutral, "temperature")

            # Use the neutral-level profiles as backup if there is only one state, otherwise no backup
            backup_ids = None if len(neutral.state) > 1 else neutral

            for state in neutral.state:
                species_data = get_neutral_state(state, elements)
                species_tuple = astuple(species_data)
                if species_tuple in neutral_bundle_uuids:
                    print(f"Warning! Skipping duplicated neutral: {species_data}")
                    continue
                neutral_bundle_uuids.add(species_tuple)

                profile_data = load_core_profiles(
                    state, species_data, backup_species_struct=backup_ids
                )

                if backup_ids is None and profile_data.temperature is None:
                    profile_data.temperature = shared_temperature

                if species_data.species_type == SpeciesType.MOLECULAR_BUNDLE:
                    composition.molecular_bundle.append(profile_data)
                elif species_data.species_type == SpeciesType.MOLECULE:
                    composition.molecule.append(profile_data)
                elif species_data.species_type == SpeciesType.NEUTRAL_BUNDLE:
                    composition.neutral_bundle.append(profile_data)
                elif species_data.species_type == SpeciesType.NEUTRAL:
                    composition.neutral.append(profile_data)
                else:
                    print(f"Warning! Skipping neutral with unexpected species {species_data}")

        # -------------------
        # === Non-bundled ===
        # -------------------
        else:
            species_data = get_neutral(neutral, elements)
            species_tuple = astuple(species_data)
            if species_tuple in neutral_uuids:
                print(f"Warning! Skipping duplicated neutral: {species_data}")
            else:
                profile_data = load_core_profiles(neutral, species_data)
                if species_data.species_type == SpeciesType.MOLECULE:
                    composition.molecule.append(profile_data)
                    neutral_uuids.add(species_tuple)
                elif species_data.species_type == SpeciesType.NEUTRAL:
                    composition.neutral.append(profile_data)
                    neutral_uuids.add(species_tuple)
                else:
                    print(
                        f"Warning! Skipping non-bundled neutral with unexpected species: "
                        f"{species_data}"
                    )

    # Replace missing species temperature with average ion temperature
    t_ion = _get_profile(profile_1d, "t_i_average")
    if t_ion is not None and len(t_ion):
        species_types = {field.name for field in fields(composition)}
        species_types.remove("electron")
        for species_type in species_types:
            for profile in getattr(composition, species_type):
                if getattr(profile, "temperature", None) is None:
                    print(f"Warning! Using average ion temperature for {profile.species}.")
                    profile.temperature = t_ion

    return composition



# Copyright 2023 Euratom
# Copyright 2023 United Kingdom Atomic Energy Authority
# Copyright 2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.
"""Module for bolometer utility functions."""

from enum import Enum

__all__ = ["CameraType", "GeometryType"]


class CameraType(Enum):
    """Enum for camera type.

    The type of the bolometer camera.

    Attributes
    ----------
    PINHOLE
        The camera is a pinhole camera.
    COLLIMATOR
        The camera is a collimator camera.
    OTHER
        The camera is of other type.
    """

    PINHOLE = 1
    COLLIMATOR = 2
    OTHER = 0

    @classmethod
    def from_value(cls, value: int) -> "CameraType":
        """Get the camera type from a value.

        Parameters
        ----------
        value
            The integer value to convert to a camera type.
            If the value is not a valid camera type, the default is `OTHER`.

        Returns
        -------
        `.CameraType`
            The corresponding `.CameraType` enum member.
        """
        if value in cls._value2member_map_:
            return cls(value)
        return cls.OTHER


class GeometryType(Enum):
    """Enum for geometry type.

    The geometry type of the bolometer foil or slit.

    Attributes
    ----------
    OUTLINE
        The geometry is defined by an outline.
    CIRCULAR
        The geometry is circular.
    RECTANGLE
        The geometry is rectangular.
    """

    OUTLINE = 1
    CIRCULAR = 2
    RECTANGLE = 3

    @classmethod
    def from_value(cls, value: int) -> "GeometryType":
        """Get the geometry type from a value.

        Parameters
        ----------
        value
            The integer value to convert to a geometry type.
            If the value is not a valid geometry type, the default is `RECTANGLE`.

        Returns
        -------
        `.GeometryType`
            The corresponding `.GeometryType` enum member.
        """
        if value in cls._value2member_map_:
            return cls(value)
        return cls.RECTANGLE



# Copyright 2023 Euratom
# Copyright 2023 United Kingdom Atomic Energy Authority
# Copyright 2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.
"""Module defining unstructured 2D mesh class and related methods."""

from __future__ import annotations

import sys
from typing import Literal

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override  # pyright: ignore[reportUnreachable]
import matplotlib.axes
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.collections import PolyCollection
from numpy.typing import ArrayLike, NDArray
from raysect.core.math.polygon import triangulate2d
from raysect.core.math.vector import Vector3D

from ..math import UnstructGridFunction2D, UnstructGridVectorFunction2D
from .base_mesh import GGDGrid

__all__ = ["UnstructGrid2D"]

ZERO_VECTOR = Vector3D(0, 0, 0)


class UnstructGrid2D(GGDGrid):
    """Unstructured 2D grid object.

    The grid cells are polygons. Vertices may be shared with neighbouring cells.

    To use Raysect's KDtree accelerator, each polygonal cell is triangulated.

    Parameters
    ----------
    vertices
        Array-like of shape ``(N, 2)`` containing coordinates of the polygon vertices.
    cells
        List of ``(N,)``-shaped arrays containing the vertex indices in clockwise or
        counterclockwise order for each polygonal cell in the list (the starting vertex must not be
        included twice).
    name
        Name of the grid, by default ``'Cells'``.
    coordinate_system
        Coordinate system of the grid, by default ``'cylindrical'``.
    """

    def __init__(
        self,
        vertices: ArrayLike,
        cells: list[ArrayLike],
        name: str = "Cells",
        coordinate_system: Literal["cylindrical", "cartesian"] = "cylindrical",
    ):
        vertices = np.asarray_chkfinite(vertices, dtype=np.float64)
        vertices.setflags(write=False)

        if vertices.ndim != 2:
            raise ValueError(
                "Attribute 'vertices' must be a 2D array-like. "
                + f"The number of dimensions in 'vertices' is {vertices.ndim}."
            )

        if vertices.shape[1] != 2:
            raise ValueError(
                "Attribute 'vertices' must be a (N, 2) array-like. "
                + f"The shape of 'vertices' is {vertices.shape}."
            )

        if not len(cells):
            raise ValueError("The list of cells must contain at least one element.")

        for cell in cells:
            if len(cell) < 3:
                raise ValueError(f"Cell {np.array2string(cell)} is not a polygon.")

        self._vertices: NDArray[np.float64] = vertices
        self._cells: tuple[ArrayLike, ...] = tuple(cells)

        super().__init__(name, 2, coordinate_system)

    @override
    def _initial_setup(self) -> None:
        self._interpolator = None

        self._num_cell: int = len(self._cells)

        x = self._vertices[:, 0]
        y = self._vertices[:, 1]

        ntri_total = 0
        for cell in self._cells:
            ntri_total += len(cell) - 2

        # Work out the extent of the mesh (both Cartesian and cylindrical).
        self._mesh_extent = {
            "xmin": x.min(),
            "xmax": x.max(),
            "ymin": y.min(),
            "ymax": y.max(),
            "rmin": x.min(),
            "rmax": x.max(),
            "zmin": y.min(),
            "zmax": y.max(),
        }

        # Triangulate cells
        self._triangles = np.empty((ntri_total, 3), dtype=np.int32)
        self._cell_to_triangle_map = np.empty((len(self._cells), 2), dtype=np.int32)
        self._triangle_to_cell_map = np.empty(ntri_total, dtype=np.int32)

        itri = 0
        for i, cell in enumerate(self._cells):
            ntri = len(cell) - 2
            if ntri == 1:
                self._triangles[i] = cell
            else:
                vert = self._vertices[cell]
                tri = triangulate2d(vert)
                self._triangles[itri : itri + ntri] = cell[tri]
            self._cell_to_triangle_map[i] = [itri, ntri]
            self._triangle_to_cell_map[itri : itri + ntri] = i
            itri += ntri

        self._triangles.setflags(write=False)
        self._cell_to_triangle_map.setflags(write=False)
        self._triangle_to_cell_map.setflags(write=False)

        # Calculate cell area and centroid
        self._cell_centre = np.empty((len(self._cells), 2), dtype=np.float64)
        self._cell_area = np.empty(len(self._cells), dtype=np.float64)

        vx = x[self._triangles]
        vy = y[self._triangles]
        area = 0.5 * np.abs(
            (vx[:, 0] - vx[:, 2]) * (vy[:, 1] - vy[:, 2])
            - (vx[:, 1] - vx[:, 2]) * (vy[:, 0] - vy[:, 2])
        )

        for i, cell in enumerate(self._cells):
            self._cell_centre[i] = self._vertices[cell].mean(0)
            i_start, ntri = self._cell_to_triangle_map[i]
            self._cell_area[i] = area[i_start : i_start + ntri].sum()

        self._cell_centre.setflags(write=False)
        self._cell_area.setflags(write=False)

        if self._coordinate_system == "cylindrical":
            self._cell_volume = 0.5 * np.pi * self._cell_centre[:, 0] * self._cell_area
            self._cell_volume.setflags(write=False)

    @property
    def vertices(self) -> NDArray[np.float64]:
        """Mesh vertex coordinates as ``(N, 2)`` array."""
        return self._vertices

    @property
    def cells(self) -> tuple[ArrayLike, ...]:
        """List of ``K`` polygonal cells."""
        return self._cells

    @property
    def triangles(self) -> NDArray[np.int32]:
        """Mesh triangles as ``(M, 3)`` array."""
        return self._triangles

    @property
    def triangle_to_cell_map(self) -> NDArray[np.int32]:
        """Array of shape ``(M,)`` mapping every triangle index to a grid cell ID."""
        return self._triangle_to_cell_map

    @property
    def cell_to_triangle_map(self) -> NDArray[np.int32]:
        """Array of shape ``(K, 2)`` mapping every grid cell index to triangle IDs.

        The first column is the index of the first triangle forming the cell.
        The second column is the number of triangles forming the cell.

        >>> itri, ntri = mesh.cell_to_triangle_map[icell]
        >>> tri_cell = mesh.triangles[itri : itri + ntri]
        """
        return self._cell_to_triangle_map

    @override
    def subset(self, indices: ArrayLike, name: str | None = None) -> UnstructGrid2D:
        """Create a subset UnstructGrid2D from this instance.

        Parameters
        ----------
        indices
            Indices of the cells of the original grid in the subset.
        name
            Name of the grid subset. Default is ``instance.name + " subset"``.

        Returns
        -------
        `.UnstructGrid2D`
            Subset instance.
        """
        grid = UnstructGrid2D.__new__(UnstructGrid2D)

        grid._name = name or self.name + " subset"
        grid._coordinate_system = self._coordinate_system
        grid._dimension = self._dimension
        grid._interpolator = None

        cells_original = tuple(
            self.cells[i] for i in indices
        )  # all cells in this subset but with original vertex indices
        cells_all = np.concatenate(
            cells_original
        )  # all vertex indices in this subset with repetitions
        vert_index, inv_index = np.unique(
            cells_all, return_inverse=True
        )  # all unique vertex indices in this subset
        grid._vertices = np.array(self.vertices[vert_index])  # vertices in this subset
        grid._vertices.setflags(write=False)

        # renumerating vertex indices
        cells = []  # and split
        i_start = 0
        for cell in cells_original:
            cells.append(inv_index[i_start : i_start + len(cell)])
            i_start += len(cell)
        grid._cells = tuple(cells)
        grid._num_cell = len(grid._cells)
        ntri_total = i_start - 2 * len(cells_original)

        # cell area and centres of this subset
        grid._cell_area = np.array(self.cell_area[indices])
        grid._cell_area.setflags(write=False)
        grid._cell_centre = np.array(self.cell_centre[indices])
        grid._cell_centre.setflags(write=False)

        # mesh extent of this subset
        xmin, ymin = grid._vertices.min(0)
        xmax, ymax = grid._vertices.max(0)
        grid._mesh_extent = {
            "xmin": xmin,
            "xmax": xmax,
            "ymin": ymin,
            "ymax": ymax,
            "rmin": xmin,
            "rmax": xmax,
            "zmin": ymin,
            "zmax": ymax,
        }

        # triangles and maps of this subset
        grid._triangles = np.empty((ntri_total, 3), dtype=np.int32)
        grid._cell_to_triangle_map = np.empty((len(cells), 2), dtype=np.int32)
        grid._triangle_to_cell_map = np.empty(ntri_total, dtype=np.int32)

        c2t_map = self.cell_to_triangle_map[indices]  # map with original triangle indices
        # maps original vertices to the subset, -1 if not in the subset
        subset_vertex_map = -1 * np.ones(self.vertices.shape[0], dtype=np.int32)
        subset_vertex_map[vert_index] = np.arange(vert_index.size, dtype=np.int32)

        itri = 0
        for i, cell in enumerate(cells):
            ntri = len(cell) - 2
            if ntri == 1:
                grid._triangles[i] = cell
            else:
                c2t = c2t_map[i]
                tri = self.triangles[c2t[0] : c2t[0] + c2t[1]]
                grid._triangles[itri : itri + ntri] = subset_vertex_map[tri]
            grid._cell_to_triangle_map[i] = [itri, ntri]
            grid._triangle_to_cell_map[itri : itri + ntri] = i
            itri += ntri

        grid._triangles.setflags(write=False)
        grid._cell_to_triangle_map.setflags(write=False)
        grid._triangle_to_cell_map.setflags(write=False)

        return grid

    @override
    def interpolator(
        self, grid_data: NDArray[np.float64], fill_value: float = 0
    ) -> UnstructGridFunction2D:
        """Return an `UnstructGridFunction2D` interpolator instance for the data defined on this grid.

        On the second and subsequent calls, the interpolator is created as an instance of the
        previously created interpolator sharing the same KDtree structure.

        Parameters
        ----------
        grid_data
            Array containing data in the grid cells.
        fill_value
            Value returned outside the grid, by default is 0.

        Returns
        -------
        `.UnstructGridFunction2D`
            Interpolator instance.
        """
        if self._interpolator is None:
            self._interpolator = UnstructGridFunction2D(
                self._vertices, self._triangles, self._triangle_to_cell_map, grid_data, fill_value
            )
            return self._interpolator

        return UnstructGridFunction2D.instance(self._interpolator, grid_data, fill_value)

    @override
    def vector_interpolator(
        self, grid_vectors: NDArray[np.float64], fill_vector: Vector3D = ZERO_VECTOR
    ) -> UnstructGridVectorFunction2D:
        """Return an `UnstructGridVectorFunction2D` interpolator instance for the vector data defined on this grid.

        On the second and subsequent calls, the interpolator is created as an instance of the
        previously created interpolator sharing the same KDtree structure.

        Parameters
        ----------
        grid_vectors
            ``(3, K)`` Array containing 3D vectors in the grid cells.
        fill_vector
            3D vector returned outside the grid.

        Returns
        -------
        `.UnstructGridVectorFunction2D`
            Interpolator instance.
        """
        if self._interpolator is None:
            self._interpolator = UnstructGridVectorFunction2D(
                self._vertices,
                self._triangles,
                self._triangle_to_cell_map,
                grid_vectors,
                fill_vector,
            )
            return self._interpolator

        return UnstructGridVectorFunction2D.instance(self._interpolator, grid_vectors, fill_vector)

    @override
    def __getstate__(self):
        """Serialize the state of the UnstructGrid2D instance for pickling.

        Returns
        -------
        Dictionary with the instance attributes.
        """
        state = {
            "name": self._name,
            "dimension": self._dimension,
            "coordinate_system": self._coordinate_system,
            "vertices": self._vertices,
            "cells": self._cells,
        }
        return state

    def __setstate__(self, state):
        """Restore the state of the UnstructGrid2D instance from the serialized state."""
        self._name = state["name"]
        self._dimension = state["dimension"]
        self._coordinate_system = state["coordinate_system"]
        self._vertices = state["vertices"]
        self._vertices.setflags(write=False)
        self._cells = state["cells"]

        self._initial_setup()

    def plot_triangle_mesh(
        self, data: ArrayLike | None = None, ax: matplotlib.axes.Axes | None = None
    ) -> matplotlib.axes.Axes:
        """Plot the triangle mesh grid geometry to a matplotlib figure.

        Parameters
        ----------
        data
            Data array defined on the polygonal mesh.
        ax
            Matplotlib axes to plot on. If None, a new figure and axes are created.

        Returns
        -------
        `~matplotlib.axes.Axes`
            The matplotlib axes with the plotted mesh.
        """
        if ax is None:
            _, ax = plt.subplots(constrained_layout=True)

        verts = self._vertices[self._triangles]
        if data is None:
            collection_mesh = PolyCollection(
                [verts], facecolor="none", edgecolor="b", linewidth=0.25
            )
        else:
            collection_mesh = PolyCollection([verts])
            collection_mesh.set_array(data[self._triangle_to_cell_map])
        ax.add_collection(collection_mesh)
        ax.set_aspect(1)
        ax.set_xlim(self._mesh_extent["xmin"], self._mesh_extent["xmax"])
        ax.set_ylim(self._mesh_extent["ymin"], self._mesh_extent["ymax"])

        if self._coordinate_system == "cartesian":
            ax.set_xlabel("X [m]")
            ax.set_ylabel("Y [m]")
        elif self._coordinate_system == "cylindrical":
            ax.set_xlabel("R [m]")
            ax.set_ylabel("Z [m]")

        return ax

    @override
    def plot_mesh(
        self,
        data: ArrayLike | None = None,
        ax: matplotlib.axes.Axes | None = None,
        **grid_styles: str | float,
    ) -> matplotlib.axes.Axes:
        """Plot the polygonal mesh grid geometry to a matplotlib figure.

        Parameters
        ----------
        data
            Data array defined on the polygonal mesh.
        ax
            Matplotlib axes to plot on. If None, a new figure and axes are created.

        Returns
        -------
        `~matplotlib.axes.Axes`
            The matplotlib axes with the plotted mesh.
        """
        if ax is None:
            _, ax = plt.subplots(constrained_layout=True)

        # Set default grid line styles if not provided
        grid_styles.setdefault("facecolor", "none")
        grid_styles.setdefault("edgecolor", "b")
        grid_styles.setdefault("linewidth", 0.25)

        verts = [self._vertices[cell] for cell in self._cells]
        if data is None:
            collection_mesh = PolyCollection(verts, **grid_styles)
        else:
            collection_mesh = PolyCollection(verts)
            collection_mesh.set_array(data)
        ax.add_collection(collection_mesh)
        ax.set_aspect(1)
        ax.set_xlim(self._mesh_extent["xmin"], self._mesh_extent["xmax"])
        ax.set_ylim(self._mesh_extent["ymin"], self._mesh_extent["ymax"])

        if self._coordinate_system == "cartesian":
            ax.set_xlabel("X [m]")
            ax.set_ylabel("Y [m]")
        elif self._coordinate_system == "cylindrical":
            ax.set_xlabel("R [m]")
            ax.set_ylabel("Z [m]")

        return ax



"""Provide functionality to create builtin IMAS sample datasets."""

import datetime

import numpy as np
from raysect.core.math import Point3D, Vector3D, rotate_z

from imas import DBEntry, IDSFactory
from imas.ids_defs import IDS_TIME_MODE_HOMOGENEOUS

try:
    import pooch

except ImportError:
    pooch = None

N_CH = 5  # Number of channels per camera
N_APERTURE = 3  # Number of apertures per channel (for collimator cameras)
N_SUBCOL = 3  # Number of subcollimators (for collimator cameras with subcollimator)
POSITION = (9.0, 0.0)  # (R, Z)
SLIT_WIDTH = 4.0e-3
SLIT_HEIGHT = 5.0e-3
FOIL_WIDTH = 1.3e-3
FOIL_HEIGHT = 3.8e-3
SLIT_SENSOR_SEPARATION = 4.0e-2
FOIL_SEPARATION = 5.08e-3
SLIT_SEPARATION = 7.5e-3
SUBCOL_SEPARATION = 1.0e-3

Y_AXIS = Vector3D(0, 1, 0)


def _bolo_data():
    """
    Create a mock bolometer IDS dataset.

    Returns
    -------
    IDSToplevel
        Mock bolometer IDS dataset.
    """
    ids = IDSFactory().new("bolometer")

    # Set properties
    ids.ids_properties.homogeneous_time = IDS_TIME_MODE_HOMOGENEOUS
    ids.ids_properties.comment = "Test bolometer IDS"
    ids.ids_properties.creation_date = datetime.date.today().isoformat()

    ids.time = [0.0]

    # Set the number of cameras
    ids.camera.resize(3)

    # ----------------------
    # === Pinhole camera ===
    # ----------------------
    camera = ids.camera[0]
    camera.name = "Pinhole Camera"
    camera.type = "pinhole"

    origin_slit = Point3D(POSITION[0], 0.0, POSITION[1])
    origin_foil = Point3D(POSITION[0] + SLIT_SENSOR_SEPARATION, 0.0, POSITION[1])
    basis_z = origin_foil.vector_to(origin_slit).normalise()
    basis_y = Y_AXIS.copy()
    basis_x = basis_y.cross(basis_z).normalise()

    camera.channel.resize(N_CH)
    for i_ch in range(N_CH):
        channel = camera.channel[i_ch]

        # Detector
        pos_foil = origin_foil + basis_x * (i_ch - (N_CH - 1) * 0.5) * FOIL_SEPARATION
        channel.detector.geometry_type = 3
        channel.detector.centre.r = np.hypot(pos_foil.x, pos_foil.y)
        channel.detector.centre.z = pos_foil.z
        channel.detector.centre.phi = np.arctan2(pos_foil.y, pos_foil.x)
        channel.detector.x1_width = FOIL_HEIGHT
        channel.detector.x2_width = FOIL_WIDTH
        for xyz in ["x", "y", "z"]:
            setattr(channel.detector.x1_unit_vector, xyz, getattr(basis_y, xyz))
            setattr(channel.detector.x2_unit_vector, xyz, getattr(basis_x, xyz))
            setattr(channel.detector.x3_unit_vector, xyz, getattr(basis_z, xyz))

        # Slit
        channel.aperture.resize(1)
        aperture = channel.aperture[0]
        aperture.geometry_type = 3
        aperture.centre.r = np.hypot(origin_slit.x, origin_slit.y)
        aperture.centre.z = origin_slit.z
        aperture.centre.phi = np.arctan2(origin_slit.y, origin_slit.x)
        aperture.x1_width = SLIT_HEIGHT
        aperture.x2_width = SLIT_WIDTH
        for xyz in ["x", "y", "z"]:
            setattr(aperture.x1_unit_vector, xyz, getattr(basis_y, xyz))
            setattr(aperture.x2_unit_vector, xyz, getattr(basis_x, xyz))
            setattr(aperture.x3_unit_vector, xyz, getattr(basis_z, xyz))

    # ---------------------------------------------
    # === Collimator camera (w/o subcollimator) ===
    # ---------------------------------------------
    camera = ids.camera[1]
    camera.name = "Collimator Camera"
    camera.type = "collimator"

    angle = 90.0  # [deg]  Angle of the collimator camera in toroidal

    origin_slit = Point3D(POSITION[0], 0.0, POSITION[1]).transform(rotate_z(angle))
    origin_foil = Point3D(POSITION[0] + SLIT_SENSOR_SEPARATION, 0.0, POSITION[1]).transform(
        rotate_z(angle)
    )
    basis_z = origin_foil.vector_to(origin_slit).normalise()
    basis_y = Y_AXIS.transform(rotate_z(angle))
    basis_x = basis_y.cross(basis_z).normalise()

    camera.channel.resize(N_CH)
    for i_ch in range(N_CH):
        channel = camera.channel[i_ch]

        # Detector
        pos_foil = origin_foil + basis_x * (i_ch - (N_CH - 1) * 0.5) * FOIL_SEPARATION
        channel.detector.geometry_type = 3
        channel.detector.centre.r = np.hypot(pos_foil.x, pos_foil.y)
        channel.detector.centre.z = pos_foil.z
        channel.detector.centre.phi = np.arctan2(pos_foil.y, pos_foil.x)
        channel.detector.x1_width = FOIL_HEIGHT
        channel.detector.x2_width = FOIL_WIDTH
        for xyz in ["x", "y", "z"]:
            setattr(channel.detector.x1_unit_vector, xyz, getattr(basis_y, xyz))
            setattr(channel.detector.x2_unit_vector, xyz, getattr(basis_x, xyz))
            setattr(channel.detector.x3_unit_vector, xyz, getattr(basis_z, xyz))

        # Slit (w/ inner apertures)
        pos_slit = origin_slit + basis_x * (i_ch - (N_CH - 1) * 0.5) * SLIT_SEPARATION
        _v = pos_foil.vector_to(pos_slit)

        channel.aperture.resize(N_APERTURE)
        for i_ap in range(N_APERTURE):
            pos_ap = pos_slit - _v * i_ap / N_APERTURE

            aperture = channel.aperture[i_ap]
            aperture.geometry_type = 3
            aperture.centre.r = np.hypot(pos_ap.x, pos_ap.y)
            aperture.centre.z = pos_ap.z
            aperture.centre.phi = np.arctan2(pos_ap.y, pos_ap.x)
            aperture.x1_width = FOIL_HEIGHT + (SLIT_HEIGHT - FOIL_HEIGHT) * (1 - i_ap / N_APERTURE)
            aperture.x2_width = FOIL_WIDTH + (SLIT_WIDTH - FOIL_WIDTH) * (1 - i_ap / N_APERTURE)
            for xyz in ["x", "y", "z"]:
                setattr(aperture.x1_unit_vector, xyz, getattr(basis_y, xyz))
                setattr(aperture.x2_unit_vector, xyz, getattr(basis_x, xyz))
                setattr(aperture.x3_unit_vector, xyz, getattr(basis_z, xyz))

    # --------------------------------------------
    # === Collimator camera (w/ subcollimator) ===
    # --------------------------------------------
    camera = ids.camera[2]
    camera.name = "Collimator Camera (w/ subcollimator)"
    camera.type = "collimator"

    angle = 180.0  # [deg]  Angle of the collimator camera in toroidal

    origin_slit = Point3D(POSITION[0], 0.0, POSITION[1]).transform(rotate_z(angle))
    origin_foil = Point3D(POSITION[0] + SLIT_SENSOR_SEPARATION, 0.0, POSITION[1]).transform(
        rotate_z(angle)
    )
    basis_z = origin_foil.vector_to(origin_slit).normalise()
    basis_y = Y_AXIS.transform(rotate_z(angle))
    basis_x = basis_y.cross(basis_z).normalise()

    camera.channel.resize(N_CH)
    for i_ch in range(N_CH):
        channel = camera.channel[i_ch]

        # Detector
        pos_foil = origin_foil + basis_x * (i_ch - (N_CH - 1) * 0.5) * FOIL_SEPARATION
        channel.detector.geometry_type = 3
        channel.detector.centre.r = np.hypot(pos_foil.x, pos_foil.y)
        channel.detector.centre.z = pos_foil.z
        channel.detector.centre.phi = np.arctan2(pos_foil.y, pos_foil.x)
        channel.detector.x1_width = FOIL_HEIGHT
        channel.detector.x2_width = FOIL_WIDTH
        for xyz in ["x", "y", "z"]:
            setattr(channel.detector.x1_unit_vector, xyz, getattr(basis_y, xyz))
            setattr(channel.detector.x2_unit_vector, xyz, getattr(basis_x, xyz))
            setattr(channel.detector.x3_unit_vector, xyz, getattr(basis_z, xyz))

        # Slit (w/ inner apertures & subcollimator)
        pos_slit = origin_slit + basis_x * (i_ch - (N_CH - 1) * 0.5) * SLIT_SEPARATION
        _v = pos_foil.vector_to(pos_slit)

        channel.subcollimators_n = N_SUBCOL
        channel.subcollimators_separation = SUBCOL_SEPARATION

        channel.aperture.resize(N_APERTURE * N_SUBCOL)
        for i_ap in range(N_APERTURE):
            pos_ap = pos_slit - _v * i_ap / N_APERTURE
            width = FOIL_WIDTH + (SLIT_WIDTH - FOIL_WIDTH) * (1 - i_ap / N_APERTURE)
            height = FOIL_HEIGHT + (SLIT_HEIGHT - FOIL_HEIGHT) * (1 - i_ap / N_APERTURE)

            for i_subcol in range(N_SUBCOL):
                pos_ap_subcol = (
                    pos_ap
                    + basis_y
                    * (i_subcol - (N_SUBCOL - 1) * 0.5)
                    * (height + SUBCOL_SEPARATION)
                    / N_SUBCOL
                )

                aperture = channel.aperture[i_ap * N_SUBCOL + i_subcol]
                aperture.geometry_type = 3
                aperture.centre.r = np.hypot(pos_ap_subcol.x, pos_ap_subcol.y)
                aperture.centre.z = pos_ap_subcol.z
                aperture.centre.phi = np.arctan2(pos_ap_subcol.y, pos_ap_subcol.x)
                aperture.x1_width = (height - SUBCOL_SEPARATION * (N_SUBCOL - 1.0)) / N_SUBCOL
                aperture.x2_width = width
                for xyz in ["x", "y", "z"]:
                    setattr(aperture.x1_unit_vector, xyz, getattr(basis_y, xyz))
                    setattr(aperture.x2_unit_vector, xyz, getattr(basis_x, xyz))
                    setattr(aperture.x3_unit_vector, xyz, getattr(basis_z, xyz))

    return ids


def bolometer_moc() -> str:
    """Return the path to a mock bolometer dataset for testing purposes.

    Returns
    -------
    str
        Path to the mock bolometer dataset file.

    Raises
    ------
    ImportError
        If the `pooch` library is not installed, which is required to fetch the dataset.

    Examples
    --------
    >>> from cherab.imas import datasets
    >>> data_path = datasets.bolometer_moc()
    >>> data_path
    '.../cherab/imas/bolometer_moc.nc'
    """
    if pooch is None:
        raise ImportError("The 'pooch' library is required to fetch the bolometer dataset.")

    path = pooch.os_cache("cherab/imas") / "bolometer_moc.nc"

    path.parent.mkdir(parents=True, exist_ok=True)

    if not path.exists():
        # Create the mock bolometer dataset and save it to the cache path
        ids = _bolo_data()
        with DBEntry(str(path), "w") as entry:
            entry.put(ids)

    return str(path)



# Copyright 2023 Euratom
# Copyright 2023 United Kingdom Atomic Energy Authority
# Copyright 2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.
"""Module for loading bolometer cameras from IMAS bolometer IDS."""

from __future__ import annotations

from itertools import cycle
from typing import Literal, overload

import numpy as np
from raysect.core.constants import ORIGIN
from raysect.core.math import AffineMatrix3D, Point3D, Vector3D, rotate_basis, translate
from raysect.core.scenegraph._nodebase import _NodeBase
from raysect.optical.loggingray import LoggingRay
from raysect.optical.material import AbsorbingSurface
from raysect.optical.scenegraph import World
from raysect.primitive import Box, Subtract, Union
from raysect.primitive.csg import CSGPrimitive

from cherab.tools.observers.bolometry import BolometerCamera, BolometerFoil, BolometerSlit
from imas.db_entry import DBEntry

from ..ids.bolometer import load_cameras
from ..ids.bolometer._camera import BoloCamera, Geometry
from ..ids.bolometer.utility import CameraType, GeometryType
from ..ids.common import get_ids_time_slice

__all__ = ["load_bolometers", "visualize"]

X_AXIS = Vector3D(1, 0, 0)
Y_AXIS = Vector3D(0, 1, 0)
Z_AXIS = Vector3D(0, 0, 1)
THICKNESS = 2.0e-3  # Thickness of camera box walls
THICKNESS_INNER_LAYER = 0.5e-3  # Thickness of inner aperture layers
EPS = 1e-4  # Small epsilon value to avoid numerical issues


@overload
def load_bolometers(
    backend_id: int,
    db_name: str,
    pulse: int,
    run: int,
    user_name: str | None = None,
    data_version: str | None = None,
    *,
    parent: _NodeBase | None = None,
    shot: int | None = None,
    dd_version: str | None = None,
    xml_path: str | None = None,
) -> list[BolometerCamera]: ...


@overload
def load_bolometers(
    uri: str,
    mode: str,
    *,
    parent: _NodeBase | None = None,
    dd_version: str | None = None,
    xml_path: str | None = None,
) -> list[BolometerCamera]: ...


def load_bolometers(
    *args,
    parent: _NodeBase | None = None,
    **kwargs,
) -> list[BolometerCamera]:
    """Load bolometer cameras from IMAS bolometer IDS.

    .. note::
        This function requires the Data dictionary v4.1.0 or later.

    Parameters
    ----------
    *args
        Arguments passed to `~imas.db_entry.DBEntry`.
    parent
        The parent node of `~cherab.tools.observers.bolometry.BolometerCamera` in the Raysect
        scene-graph, by default None.
    **kwargs
        Keyword arguments passed to `~imas.db_entry.DBEntry` constructor.

    Returns
    -------
    `list[BolometerCamera]`
        List of `~cherab.tools.observers.bolometry.BolometerCamera` objects.

    Raises
    ------
    ValueError
        If slit data is required for a pinhole camera but not provided.
    NotImplementedError
        If an unsupported camera type or geometry type is encountered.

    Examples
    --------
    >>> from raysect.optical import World
    >>> world = World()

    If you have a local IMAS database and store the "bolometer.h5" file there:

    >>> bolometers = load_bolometers("imas:hdf5?path=path/to/db/", "r", parent=world)

    If you want to load netCDF files directly:

    >>> bolometers = load_bolometers("path/to/bolometer_file.nc", "r", parent=world)
    """
    # Load bolometer IDS
    with DBEntry(*args, **kwargs) as entry:
        # Get available time slices
        ids = get_ids_time_slice(entry, "bolometer")

    # Extract bolometer data
    bolo_data = load_cameras(ids)

    bolometers: list[BolometerCamera] = []

    for data in bolo_data:
        # Skip empty cameras
        if len(data.channels) == 0:
            continue

        # ------------------
        # === Camera Box ===
        # ------------------
        camera_box = _create_camera_box(data)
        camera = BolometerCamera(camera_geometry=camera_box, name=data.name, parent=parent)

        match data.type:
            case CameraType.PINHOLE:
                # ----------------------
                # === Slit (Pinhole) ===
                # ----------------------
                # Pick up only first aperture and use it for all channels
                slit_data = data.channels[0].slits[0]
                slit = BolometerSlit(
                    f"slit-{data.name}",
                    slit_data.centre,
                    slit_data.basis_x,
                    slit_data.dx,
                    slit_data.basis_y,
                    slit_data.dy,
                    curvature_radius=(slit_data.radius or 0.0)
                    if slit_data.type == GeometryType.CIRCULAR
                    else 0.0,
                    parent=camera,
                )
            case CameraType.COLLIMATOR:
                slit = None  # Defined per channel below
            case _:
                raise NotImplementedError(f"Camera type {data.type} not supported yet.")

        for i_channel, channel in enumerate(data.channels):
            # -------------------------
            # === Slit (Collimator) ===
            # -------------------------
            # Concatenate top plate slits into one slit if multiple sub-collimators
            # NOTE: Top plate slits are accumulated in the beginning of the slits list
            # NOTE: All slit geometry types are assumed to be the RECTANGLE.
            # NOTE: Sub-collimators are assumed to be aligned in a row along the basis_y direction.
            if (num_subcol := channel.num_subcollimator) > 1:
                _slit_data = channel.slits[:num_subcol]
                centre = (
                    ORIGIN
                    + sum([ORIGIN.vector_to(s.centre) for s in _slit_data], Vector3D(0, 0, 0))
                    / num_subcol
                )
                basis_x = _slit_data[0].basis_x
                basis_y = _slit_data[0].basis_y
                dx = _slit_data[0].dx
                dy = (
                    ORIGIN
                    + ORIGIN.vector_to(_slit_data[0].centre)
                    - _slit_data[0].dy * 0.5 * basis_y
                ).distance_to(
                    ORIGIN
                    + ORIGIN.vector_to(_slit_data[-1].centre)
                    + _slit_data[-1].dy * 0.5 * basis_y
                )

                slit_data = Geometry(
                    type=GeometryType.RECTANGLE,
                    centre=centre,
                    basis_x=basis_x,
                    basis_y=basis_y,
                    dx=dx,
                    dy=dy,
                )
            else:
                slit_data = channel.slits[0]

            if data.type == CameraType.COLLIMATOR:
                # Create slit object
                match slit_data.type:
                    case GeometryType.CIRCULAR | GeometryType.RECTANGLE:
                        slit = BolometerSlit(
                            f"slit-{data.name}-ch{i_channel}",
                            slit_data.centre,
                            slit_data.basis_x,
                            slit_data.dx,
                            slit_data.basis_y,
                            slit_data.dy,
                            curvature_radius=(slit_data.radius or 0.0)
                            if slit_data.type == GeometryType.CIRCULAR
                            else 0.0,
                            parent=camera,
                        )
                    case _:
                        raise NotImplementedError("Outline geometry not supported yet.")
            else:
                if slit is None:
                    raise ValueError("Slit data is required for pinhole camera.")

            # ------------
            # === Foil ===
            # ------------
            match channel.foil.type:
                case GeometryType.CIRCULAR | GeometryType.RECTANGLE:
                    foil = BolometerFoil(
                        f"foil-{data.name}-ch{i_channel}",
                        channel.foil.centre,
                        channel.foil.basis_x,
                        channel.foil.dx,
                        channel.foil.basis_y,
                        channel.foil.dy,
                        slit,
                        curvature_radius=(channel.foil.radius or 0.0)
                        if channel.foil.type == GeometryType.CIRCULAR
                        else 0.0,
                        parent=camera,
                    )
                case _:
                    raise NotImplementedError("Outline geometry not supported yet.")

            # Add foil to the camera
            camera.add_foil_detector(foil)

        bolometers.append(camera)

    return bolometers


def _create_camera_box(bolo_data: BoloCamera) -> CSGPrimitive:
    """Create a camera housing box geometry.

    This box is represented as a Subtract primitive, where a smaller inner box is subtracted
    from a larger outer box to create a hollow structure.
    Additionally, top plate apertures and other apertures (inner layers) are included as Unions.

    Returns
    -------
    `CSGPrimitive`
        The camera housing box geometry.
    """
    # Extract slits data
    slits_top: list[Geometry]
    slits_inner: list[list[Geometry]] = []
    if bolo_data.type == CameraType.PINHOLE:
        # Pick up only first aperture and use it for all channels
        slits_top = [bolo_data[0].slits[0]]
    else:
        # Consider sub-collimators for collimator cameras
        num_subcol = bolo_data[0].num_subcollimator
        num_channel = len(bolo_data)
        _s: list[Geometry] = []
        slits_top = sum(
            [bolo_data[i_ch].slits[:num_subcol] for i_ch in range(num_channel)],
            _s,
        )
        slits_inner = [
            sum(
                [
                    bolo_data[i_ch].slits[i_slit : i_slit + num_subcol]
                    for i_ch in range(num_channel)
                ],
                _s,
            )
            for i_slit in range(num_subcol, len(bolo_data[0].slits), num_subcol)
        ]

    # -------------------------------
    # === Local coordinate system ===
    # -------------------------------
    # The local origin is placed at the average position of top plate slits of all foil detectors,
    # and the basis vectors are defined based on the average normal and basis_x vectors of the slits.
    origin = ORIGIN + sum(
        [ORIGIN.vector_to(slit.centre) for slit in slits_top],
        Vector3D(0, 0, 0),
    ) / len(slits_top)

    basis_z: Vector3D = sum(
        [slit.basis_z for slit in slits_top],
        Vector3D(0, 0, 0),
    ).normalise()

    basis_x: Vector3D = sum(
        [slit.basis_x for slit in slits_top],
        Vector3D(0, 0, 0),
    ).normalise()

    basis_y = basis_z.cross(basis_x).normalise()

    # Transformation matrix from local to global coordinate system
    to_root = translate(*origin) * rotate_basis(basis_z, basis_y)

    # ------------------------------
    # === Determine box geometry ===
    # ------------------------------
    EPS_WIDTH = 1e-3
    EPS_HEIGHT = 1e-3
    EPS_DEPTH = 1e-3

    # Collect all corner points of foils and slits
    pts: list[Point3D] = []
    for geometry in [bolo_data[i].foil for i in range(len(bolo_data))] + slits_top:
        pts += _get_verts(geometry)

    # Camera box dimensions (in local coordinate system)
    box_width_u = max([origin.vector_to(p).dot(basis_x) for p in pts]) + EPS_WIDTH
    box_width_l = min([origin.vector_to(p).dot(basis_x) for p in pts]) - EPS_WIDTH
    box_height_u = max([origin.vector_to(p).dot(basis_y) for p in pts]) + EPS_HEIGHT
    box_height_l = min([origin.vector_to(p).dot(basis_y) for p in pts]) - EPS_HEIGHT
    box_depth_u = max([origin.vector_to(p).dot(basis_z) for p in pts])  # slit top plate
    box_depth_l = min([origin.vector_to(p).dot(basis_z) for p in pts]) - EPS_DEPTH

    # -------------------------
    # === Create Hollow Box ===
    # -------------------------
    camera_box = Box(
        lower=ORIGIN + box_width_l * X_AXIS + box_height_l * Y_AXIS + box_depth_l * Z_AXIS,
        upper=ORIGIN + box_width_u * X_AXIS + box_height_u * Y_AXIS + box_depth_u * Z_AXIS,
        name=f"inner-box-{bolo_data.name}",
    )
    outside_box = Box(
        lower=camera_box.lower - Vector3D(THICKNESS, THICKNESS, THICKNESS),
        upper=camera_box.upper + Vector3D(THICKNESS, THICKNESS, THICKNESS),
        name=f"outside-box-{bolo_data.name}",
    )
    camera_box = Subtract(outside_box, camera_box, name=f"hollow-box-{bolo_data.name}")

    # ------------------------------------------
    # === Clip out Top Plate Apertures Holes ===
    # ------------------------------------------
    for slit in slits_top:
        slit_verts = _get_verts(slit)
        aperture_box = Box(
            lower=ORIGIN
            + min([origin.vector_to(p).dot(basis_x) for p in slit_verts]) * X_AXIS
            + min([origin.vector_to(p).dot(basis_y) for p in slit_verts]) * Y_AXIS
            + (box_depth_u - EPS) * Z_AXIS,
            upper=ORIGIN
            + max([origin.vector_to(p).dot(basis_x) for p in slit_verts]) * X_AXIS
            + max([origin.vector_to(p).dot(basis_y) for p in slit_verts]) * Y_AXIS
            + (box_depth_u + THICKNESS + EPS) * Z_AXIS,
            name=f"aperture-box-top-{bolo_data.name}",
        )
        camera_box = Subtract(camera_box, aperture_box, name=f"aperture-top-{bolo_data.name}")

    # ----------------------------------
    # === Add Inner Apertures Plates ===
    # ----------------------------------
    # NOTE: Assume all inner apertures use the same local coordinate system as the top plate.
    for slits in slits_inner:
        # Create Inner Aperture Plate layer
        layer_depth_z = basis_z.dot(origin.vector_to(slits[0].centre))
        layer = Box(
            lower=Point3D(
                outside_box.lower.x,
                outside_box.lower.y,
                layer_depth_z - THICKNESS_INNER_LAYER * 0.5,
            ),
            upper=Point3D(
                outside_box.upper.x,
                outside_box.upper.y,
                layer_depth_z + THICKNESS_INNER_LAYER * 0.5,
            ),
            name=f"inner-plate-{bolo_data.name}",
        )

        for slit in slits:
            slit_verts = _get_verts(slit)
            aperture_box = Box(
                lower=ORIGIN
                + min([origin.vector_to(p).dot(basis_x) for p in slit_verts]) * X_AXIS
                + min([origin.vector_to(p).dot(basis_y) for p in slit_verts]) * Y_AXIS
                + (layer_depth_z - THICKNESS_INNER_LAYER * 0.5 - EPS) * Z_AXIS,
                upper=ORIGIN
                + max([origin.vector_to(p).dot(basis_x) for p in slit_verts]) * X_AXIS
                + max([origin.vector_to(p).dot(basis_y) for p in slit_verts]) * Y_AXIS
                + (layer_depth_z + THICKNESS_INNER_LAYER * 0.5 + EPS) * Z_AXIS,
                name=f"inner-aperture-box-{bolo_data.name}",
            )
            layer = Subtract(layer, aperture_box, name=f"inner-aperture-plate-{bolo_data.name}")

        camera_box = Union(camera_box, layer)

    # Transform to global coordinate system
    camera_box.transform = to_root

    # Apply absorbing material
    camera_box.material = AbsorbingSurface()

    # Name the camera box
    camera_box.name = f"camera-box-{bolo_data.name}"

    return camera_box


def _get_verts(geometry: Geometry) -> list[Point3D]:
    """Get the geometry vertices.

    For example, if the geometry is rectangular, return 4 corner vertices.

    Parameters
    ----------
    geometry
        Geometry structure object.

    Returns
    -------
    `list[Point3D]`
        List of geometry vertices.
    """
    dx = geometry.dx
    dy = geometry.dy
    basis_x = geometry.basis_x
    basis_y = geometry.basis_y
    center = geometry.centre
    match geometry.type:
        case GeometryType.RECTANGLE:
            verts = [
                center + 0.5 * dx * basis_x + 0.5 * dy * basis_y,
                center - 0.5 * dx * basis_x + 0.5 * dy * basis_y,
                center - 0.5 * dx * basis_x - 0.5 * dy * basis_y,
                center + 0.5 * dx * basis_x - 0.5 * dy * basis_y,
            ]
        case _:
            raise NotImplementedError(f"Geometry type {geometry.type} not implemented yet.")

    return verts


def visualize(
    camera: BolometerCamera,
    fig=None,
    num_rays: int | None = None,
    ray_from_channel: int | list[int] | None = None,
    ray_terminate_distance: float = 5.0e-2,
    aspect: Literal["data", "auto"] = "data",
    show: bool = True,
):
    """Visualize a bolometer camera geometry in 3D using Plotly.

    .. note::
        This function requires `plotly` to be installed.

    Parameters
    ----------
    camera
        The BolometerCamera object to visualize.
    fig : plotly.graph_objects.Figure | None
        An existing Plotly figure to add the camera visualization to, by default None.
    num_rays
        Number of rays to trace for visualizing the field of view, by default None.
    ray_from_channel
        Channel index or list of channel indices from which to trace rays, by default None.
    ray_terminate_distance
        Distance at which rays are terminated, by default 5.0e-2.
    aspect
        Aspect ratio of the plot, by default "data".
    show
        Whether to display the plot, by default True.

    Returns
    -------
    `plotly.graph_objects.Figure`
        The Plotly figure object.

    Raises
    ------
    ImportError
        If Plotly is not installed.
    TypeError
        If the provided fig is not a Plotly Figure instance.

    Examples
    --------
    >>> bolometers = load_bolometers("imas:hdf5?path=path/to/db/", "r")
    >>> fig = visualize(bolometers[0], num_rays=100, ray_from_channel=[0, 3])
    """
    try:
        from plotly import graph_objects as go
        from plotly.colors import qualitative
    except ImportError as e:
        raise ImportError("Plotly is required for visualization.") from e

    if fig is None:
        fig = go.Figure()
    else:
        if not isinstance(fig, go.Figure):
            raise TypeError("fig must be a plotly.graph_objects.Figure instance.")

    # Set up rays to be traced
    foils_ray_triggered: list[BolometerFoil] = []
    if isinstance(num_rays, int) and num_rays > 0:
        if isinstance(ray_from_channel, int):
            foils_ray_triggered = [camera.foil_detectors[ray_from_channel]]
        elif isinstance(ray_from_channel, list):
            foils_ray_triggered = [camera.foil_detectors[ch] for ch in ray_from_channel]
        else:
            foils_ray_triggered = camera.foil_detectors

    # Create scene graph temporally
    world = World()
    prev_parent = camera.parent
    try:
        camera.parent = world

        # === Local Axis ===
        local_origin = ORIGIN + sum(
            [ORIGIN.vector_to(foil.slit.centre_point) for foil in camera.foil_detectors],
            Vector3D(0, 0, 0),
        ) / len(camera.foil_detectors)
        local_z_axis: Vector3D = sum(
            [foil.slit.normal_vector.normalise() for foil in camera.foil_detectors],
            Vector3D(0, 0, 0),
        ).normalise()
        local_x_axis: Vector3D = sum(
            [foil.slit.basis_x.normalise() for foil in camera.foil_detectors],
            Vector3D(0, 0, 0),
        ).normalise()

        local_y_axis = local_z_axis.cross(local_x_axis).normalise()

        # --------------------------
        # === Plot Slits & Foils ===
        # --------------------------
        slit_name = ""
        for i_ch, foil in enumerate(camera.foil_detectors):
            foil: BolometerFoil
            for rect in [foil.slit, foil]:
                center: Point3D = rect.centre_point

                if isinstance(rect, BolometerSlit):
                    if rect.name == slit_name:
                        continue
                    else:
                        slit_name = rect.name
                        dx = rect.dx
                        dy = rect.dy
                elif isinstance(rect, BolometerFoil):
                    dx = rect.x_width
                    dy = rect.y_width
                else:
                    raise TypeError("Unknown rectangle type")

                basis_x: Vector3D = rect.basis_x.normalise()
                basis_y: Vector3D = rect.basis_y.normalise()

                # Calculate the foil edge coordinates
                vertices = [
                    center + 0.5 * dx * basis_x + 0.5 * dy * basis_y,
                    center - 0.5 * dx * basis_x + 0.5 * dy * basis_y,
                    center - 0.5 * dx * basis_x - 0.5 * dy * basis_y,
                    center + 0.5 * dx * basis_x - 0.5 * dy * basis_y,
                    center + 0.5 * dx * basis_x + 0.5 * dy * basis_y,  # Close the loop
                ]
                color = "blue" if rect is foil else "green"
                fig.add_trace(
                    go.Scatter3d(
                        x=[v.x for v in vertices],
                        y=[v.y for v in vertices],
                        z=[v.z for v in vertices],
                        mode="lines",
                        name=f"{'Foil' if rect is foil else 'Slit'} - {i_ch} CH",
                        line=dict(color=color, width=3),
                        hovertemplate=f"Width: {dx * 1e3:.2f} mm<br>Height: {dy * 1e3:.2f} mm",
                    )
                )

        # ------------------------------
        # === Camera Box (Outer Box) ===
        # ------------------------------
        if isinstance(camera._camera_geometry, CSGPrimitive):
            boxes = _extract_box_primitive(camera._camera_geometry)
            if boxes:
                for box in boxes:
                    fig: go.Figure = _plot_box(
                        box, transform=camera._camera_geometry.transform, fig=fig
                    )

        # -----------------
        # === Plot Rays ===
        # -----------------
        text_num_rays_passed = ""
        if isinstance(num_rays, int) and num_rays > 0:
            # Add terminating box to avoid rays escaping to infinity
            terminate_board = Box(
                lower=Point3D(-1e9, -1e9, -1e-3),
                upper=Point3D(1e9, 1e9, 0),
                parent=world,
                name="terminating_board",
            )
            terminate_board.material = AbsorbingSurface()
            terminate_board.transform = (
                translate(*local_origin)
                * rotate_basis(local_z_axis, local_y_axis)
                * translate(0, 0, ray_terminate_distance)
            )

            # Generate and trace rays
            count = 0
            for foil, color in zip(foils_ray_triggered, cycle(qualitative.Set1), strict=False):
                for ray in foil._generate_rays(LoggingRay(), num_rays):
                    origin = ray[0].origin.transform(foil.to_root())
                    direction = ray[0].direction.transform(foil.to_root())
                    ray = LoggingRay(origin=origin, direction=direction)
                    ray.trace(world)
                    hit_points = [origin, ray.path_vertices[-1]]
                    if ray.log[-1].primitive.name == "terminating_board":
                        count += 1
                    fig.add_trace(
                        go.Scatter3d(
                            x=[p.x for p in hit_points],
                            y=[p.y for p in hit_points],
                            z=[p.z for p in hit_points],
                            mode="lines+markers",
                            marker=dict(size=1.5, color=color),
                            line=dict(color=color, width=1),
                            name="rays",
                            showlegend=False,
                            hovertemplate=(
                                f"From: {foil.name}<br>"
                                f"Hit Object: {ray.log[-1].primitive.name}<extra></extra>"
                            ),
                        )
                    )
            total_rays = num_rays * len(foils_ray_triggered)
            if total_rays > 0:
                text_num_rays_passed = f" ({count / total_rays:.2%} Rays Passed)"

        # -----------------------
        # === Plot local axes ===
        # -----------------------
        scale = 0.01
        axes = [
            (local_x_axis, "X Axis", "rgb(255, 0, 0)"),
            (local_y_axis, "Y Axis", "rgb(0, 255, 0)"),
            (local_z_axis, "Z Axis", "rgb(0, 0, 255)"),
        ]
        for axis, name, color in axes:
            point = local_origin + scale * axis
            fig.add_trace(
                go.Scatter3d(
                    x=[local_origin.x, point.x],
                    y=[local_origin.y, point.y],
                    z=[local_origin.z, point.z],
                    name=name,
                    marker=dict(color=color, size=2),
                    line=dict(color=color),
                    showlegend=False,
                )
            )

        fig.update_layout(
            title=f"Bolometer Camera: {camera.name}{text_num_rays_passed}",
            scene=dict(
                xaxis_title="X (m)",
                yaxis_title="Y (m)",
                zaxis_title="Z (m)",
                aspectmode=aspect,
            ),
            showlegend=True,
            width=700,
            height=600,
            margin=dict(r=10, l=10, b=10, t=35),
        )

        # restore previous parent node
        camera.parent = prev_parent

        if show:
            fig.show()

    finally:
        camera.parent = prev_parent

    return fig


def _extract_box_primitive(primitive: CSGPrimitive) -> list[Box]:
    """Recursively extract all Box primitives from CSGPrimitive.

    Parameters
    ----------
    primitive
        CSGPrimitive object to extract Box primitives from.

    Returns
    -------
    `list[Box]`
        List of Box primitives found within the CSGPrimitive.
    """
    boxes = []

    # Check primitive_a
    primitive_a = getattr(primitive, "primitive_a", None)
    if primitive_a is not None:
        if isinstance(primitive_a, Box):
            boxes.append(primitive_a)
        elif isinstance(primitive_a, CSGPrimitive):
            boxes.extend(_extract_box_primitive(primitive_a))

    # Check primitive_b
    primitive_b = getattr(primitive, "primitive_b", None)
    if primitive_b is not None:
        if isinstance(primitive_b, Box):
            boxes.append(primitive_b)
        elif isinstance(primitive_b, CSGPrimitive):
            boxes.extend(_extract_box_primitive(primitive_b))

    return boxes


def _plot_box(
    box: Box,
    transform: AffineMatrix3D,
    fig,
    color: str = "#7d7d7d",
):
    """Plot a box in a given figure.

    Parameters
    ----------
    box
        Box primitive to plot.
    transform
        Affine transformation to apply to the box vertices.
    fig
        Plotly figure to add the box plot to.
    color
        Color of the box, by default "#7d7d7d".

    Returns
    -------
    `plotly.graph_objects.Figure`
        The updated Plotly figure with the box plotted.
    """
    from plotly import graph_objects as go

    # Get box vertices
    _xaxis = X_AXIS.transform(transform)
    _yaxis = Y_AXIS.transform(transform)
    _zaxis = Z_AXIS.transform(transform)
    lower = box.lower.transform(transform)
    upper = box.upper.transform(transform)
    lower_to_upper = lower.vector_to(upper)
    box_width = abs(lower_to_upper.dot(_xaxis))
    box_height = abs(lower_to_upper.dot(_yaxis))
    box_depth = abs(lower_to_upper.dot(_zaxis))
    vertices = [
        lower,
        lower + box_width * _xaxis,
        lower + box_width * _xaxis + box_height * _yaxis,
        lower + box_height * _yaxis,
        lower + box_depth * _zaxis,
        lower + box_depth * _zaxis + box_width * _xaxis,
        upper,
        upper - box_width * _xaxis,
    ]
    verts = np.array([[*vertex] for vertex in vertices])

    # Plot box surfaces
    fig.add_trace(
        go.Mesh3d(
            x=verts[:, 0],
            y=verts[:, 1],
            z=verts[:, 2],
            i=[7, 0, 0, 0, 4, 4, 6, 6, 4, 0, 3, 2],
            j=[3, 4, 1, 2, 5, 6, 5, 2, 0, 1, 6, 3],
            k=[0, 7, 2, 3, 6, 7, 1, 1, 5, 5, 7, 6],
            opacity=0.2,
            color=color,
            flatshading=True,
            name="Camera Box",
            showlegend=True,
        )
    )

    # Plot box edges
    # Define the 12 edges of the box
    edges = [
        # bottom face
        (0, 1),
        (1, 2),
        (2, 3),
        (3, 0),
        # top face
        (4, 5),
        (5, 6),
        (6, 7),
        (7, 4),
        # vertical edges
        (0, 4),
        (1, 5),
        (2, 6),
        (3, 7),
    ]

    # Plot edges as lines
    for edge in edges:
        v1, v2 = vertices[edge[0]], vertices[edge[1]]
        fig.add_trace(
            go.Scatter3d(
                x=[v1.x, v2.x],
                y=[v1.y, v2.y],
                z=[v1.z, v2.z],
                mode="lines",
                line=dict(color=color, width=2),
                name="Camera Box Edge",
                showlegend=False,
                hoverinfo="skip",
            )
        )

    return fig



# Copyright 2023 Euratom
# Copyright 2023 United Kingdom Atomic Energy Authority
# Copyright 2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.
"""Module for loading wall components from wall IDSs."""

import numpy as np
from raysect.core.math.polygon import triangulate2d

from imas.ids_structure import IDSStructure

__all__ = ["load_wall_3d"]

VERTEX_DIMENSION = 0
POLYGON_DIMENSION = 2


# TODO: Check coordinate types and convert to Cartesian if required
def load_wall_3d(
    description_ggd: IDSStructure, subsets: list[str] | None = None
) -> dict[str, dict[str, np.ndarray]]:
    """Load machine wall components from IMAS wall IDS.

    Parameters
    ----------
    description_ggd
        A description_ggd structure from the 'wall' IDS.
    subsets
        List of names of specific ggd subsets to load, by default None (loads all subsets).

    Returns
    -------
    `dict[str, dict[str, numpy.ndarray]]`
        Dictionary of wall components defined by vertices and triangles.
        The dictionary keys for components are assigns as follows:
        ``"{grid_name}.{subset_name}.{material_name}"``.
        So, the returning dictionary looks like:

        .. code-block:: python

            {
                "FullTokamak.full_main_chamber_wall.Be": {
                    "vertices": np.ndarray of shape (N, 3),
                    "triangles": np.ndarray of shape (M, 3),
                },
                ...
            }

    Raises
    ------
    RuntimeError
        If the grid_ggd AOS is empty in the given description_ggd.
    """
    if not len(description_ggd.grid_ggd):
        raise RuntimeError("The grid_ggd AOS is empty in the given description_ggd.")

    # get the grid
    grid = description_ggd.grid_ggd[0]

    # get the material names and the element indices of grid subsets
    material_struct = description_ggd.material
    materials = _get_materials(material_struct)

    # get the coordinates of the vertices
    space = grid.space[0]  # wall grid has only a single space (unstructured grid)
    num_vert = len(space.objects_per_dimension[VERTEX_DIMENSION].object)
    vertices = np.zeros((num_vert, 3))
    for i in range(num_vert):
        vertices[i] = space.objects_per_dimension[VERTEX_DIMENSION].object[i].geometry

    poly_in_subsets = set()  # the polygons in grid subsets
    components = {}

    grid_name = grid.identifier.name

    # iterate over the grid subsets to get individual wall components and their materials
    for subset in grid.grid_subset:
        subset_name = subset.identifier.name
        subset_index = subset.identifier.index

        if subset.dimension - 1 != POLYGON_DIMENSION:
            continue
        if subsets and (subset_name not in subsets):
            continue

        if subset_index in materials:
            # materials are specified for this subset
            for material_name, element_indices in materials[subset_index]:
                component_name = f"{grid_name}.{subset_name}.{material_name}"
                vert, tri = _get_mesh_from_subset(
                    grid, subset, element_indices, vertices, poly_in_subsets
                )
                components[component_name] = {"vertices": vert, "triangles": tri}
        else:
            # materials are not specified
            component_name = f"{grid_name}.{subset_name}.none"
            element_indices = range(len(subset.element)) if len(subset.element) else [0]
            vert, tri = _get_mesh_from_subset(
                grid, subset, element_indices, vertices, poly_in_subsets
            )
            components[component_name] = {"vertices": vert, "triangles": tri}

    # add all remaining polygons to a dedicated component
    if not subsets:
        vert, tri = _get_mesh_from_remaining_polygons(grid, vertices, poly_in_subsets)
        component_name = f"{grid_name}.none.none"
        components[component_name] = {"vertices": vert, "triangles": tri}

    return components


def _get_materials(material_struct):
    materials = {}

    if len(material_struct):  # check if the material structure is defined in this IDS
        for subset in material_struct[0].grid_subset:
            names = np.asarray(subset.identifiers.names)
            # get unique materials
            unique_names = np.unique(names)
            material_list = []
            for name in unique_names:
                (elements,) = np.where(names == name)
                material_list.append((name, elements))
            materials[subset.grid_subset_index] = material_list

    return materials


def _get_mesh_from_subset(grid, subset, element_indices, vertices, poly_in_subsets):
    space = grid.space[0]  # wall grid has only a single space (unstructured grid)
    triangles = []

    if len(subset.element):  # complex subset
        indices = [
            subset.element[iel].object[0].index - 1 for iel in element_indices
        ]  # indexing: 1 -> 0
    else:  # trivial subset, includes all polygons, element index equals to polygon index
        if len(element_indices) == 1 and element_indices[0] == 0:  # trivial case, use all polygons
            indices = range(len(space.objects_per_dimension[POLYGON_DIMENSION].object))
        else:
            indices = element_indices

    for i in indices:
        polygon = (
            space.objects_per_dimension[POLYGON_DIMENSION].object[i].nodes - 1
        )  # indexing: 1 -> 0
        if len(polygon) == 3:
            triangles.append(polygon)
        elif len(polygon) > 3:
            tri_array = polygon[_triangulate_polygon(vertices[polygon])]
            triangles += [tri for tri in tri_array]
        else:
            raise RuntimeError(f"Not a polygon: {np.array2string(polygon)}.")

    poly_in_subsets.update(indices)

    triangles = np.array(triangles, dtype=np.int32)
    vert_index, inv_index = np.unique(triangles, return_inverse=True)
    vert = vertices[vert_index]  # vertices in this subset
    tri = np.arange(len(vert), dtype=np.int32)[inv_index].reshape(triangles.shape)  # renumerate

    return vert, tri


def _get_mesh_from_remaining_polygons(grid, vertices, poly_in_subsets):
    space = grid.space[0]  # wall grid has only a single space (unstructured grid)
    num_poly = len(space.objects_per_dimension[POLYGON_DIMENSION].object)
    mask = np.ones(num_poly, dtype=bool)
    mask[list(poly_in_subsets)] = False
    (poly_remain,) = np.where(mask)

    triangles = []
    for i in poly_remain:
        polygon = (
            space.objects_per_dimension[POLYGON_DIMENSION].object[i].nodes - 1
        )  # indexing: 1 -> 0
        if len(polygon) == 3:
            triangles.append(polygon)
        elif len(polygon) > 3:
            tri_array = polygon[_triangulate_polygon(vertices[polygon])]
            triangles += [tri for tri in tri_array]
        else:
            raise RuntimeError(f"Not a polygon: {np.array2string(polygon)}.")

    triangles = np.array(triangles, dtype=np.int32)
    vert_index, inv_index = np.unique(triangles, return_inverse=True)
    vert = vertices[vert_index]  # vertices in this subset
    tri = np.arange(len(vert), dtype=np.int32)[inv_index].reshape(triangles.shape)  # renumerate

    return vert, tri


def _triangulate_polygon(vert):
    # convert to 2d coordinates
    vert -= vert[0]
    e1 = np.copy(vert[1])
    e2 = np.copy(vert[2])
    e3 = np.cross(e1, e2)
    e3 /= np.sqrt(e3 @ e3)
    e1 /= np.sqrt(e1 @ e1)
    e2 = np.cross(e3, e1)
    transform = np.linalg.inv(np.vstack((e1, e2, e3)))

    polygon2d = (vert @ transform)[:, :2]

    triangles = triangulate2d(polygon2d)

    return triangles[:, ::-1]  # clockwise -> anti-clockwise



r"""Sample Dataset utilities.

Usage of Datasets
=================

CHERAB-IMAS dataset methods can be simply called as follows: ``'<dataset-name>()'``
This downloads/creates the dataset files over the network once, and saves the cache,
before returning the path to the downloaded/created data file.

How dataset retrieval and storage works
=======================================

The dataset files are stored in public repositories on
Zenodo: https://doi.org/10.5281/zenodo.17062700.
The `cherab.imas.datasets` submodule utilizes
and depends on `Pooch <https://www.fatiando.org/pooch/latest/>`_, a Python
package built to simplify fetching data files. Pooch uses these repos to
retrieve the respective dataset files when calling the dataset function.

Datasets available in CHERAB-IMAS are sample datasets resulting from ITER JINTRAC, SOLPS, JOREK,
etc. simulations, and are intended for testing and demonstration purposes.

A registry of all the datasets, essentially a mapping of filenames with their
MD5 hash are maintained, which Pooch uses to handle and verify
the downloads on function call. After downloading the dataset once, the files
are saved in the system cache directory under ``'cherab/imas'``.

Dataset cache locations may vary on different platforms.

For macOS::

    "~/Library/Caches/cherab/imas"

For Linux and other Unix-like platforms::

    "~/.cache/cherab/imas"  # or the value of the XDG_CACHE_HOME env var, if defined

For Windows::

    "C:\Users\<user>\AppData\Local\<AppAuthor>\cherab\imas\Cache"

In environments with constrained network connectivity for various security
reasons or on systems without continuous internet connections, one may manually
load the cache of the datasets by placing the contents of the dataset repo in
the above mentioned cache directory to avoid fetching dataset errors without
the internet connectivity.
"""

from ._builtin import bolometer_moc
from ._fetchers import iter_jintrac, iter_jorek, iter_solps
from ._utils import clear_cache

__all__ = [
    "iter_jintrac",
    "iter_solps",
    "iter_jorek",
    "bolometer_moc",
    "clear_cache",
]
